@@ -1,6 +1,6 @@
 /* -*- mode: C++; c-basic-offset: 4; tab-width: 4 -*- 
  *
- * Copyright (c) 2006 Apple Computer, Inc. All rights reserved.
+ * Copyright (c) 2006-2011 Apple Inc. All rights reserved.
  *
  * @APPLE_LICENSE_HEADER_START@
  * 
@@ -27,6 +27,7 @@
 
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <sys/errno.h>
 #include <sys/mman.h>
 #include <mach/mach.h>
 #include <limits.h>
@@ -37,10 +38,11 @@
 #include <unistd.h>
 #include <mach-o/loader.h>
 #include <mach-o/fat.h>
+#include <rootless.h>
 
 #include <vector>
 #include <set>
-#include <ext/hash_map>
+#include <unordered_map>
 
 #include "MachOFileAbstraction.hpp"
 #include "Architectures.hpp"
@@ -68,11 +70,13 @@ public:
 	struct Segment
 	{
 	public:
-					Segment(uint64_t addr, uint64_t vmsize, uint64_t offset, uint64_t file_size, 
-							uint32_t prot, const char* segName) : fOrigAddress(addr), fOrigSize(vmsize),
+					Segment(uint64_t addr, uint64_t vmsize, uint64_t offset, uint64_t file_size, uint64_t sectionsSize,
+							uint64_t sectionsAlignment, uint64_t align,
+							uint32_t prot, uint32_t sectionCount, const char* segName) : fOrigAddress(addr), fOrigSize(vmsize),
 							fOrigFileOffset(offset),  fOrigFileSize(file_size), fOrigPermissions(prot), 
-							fSize(vmsize), fFileOffset(offset), fFileSize(file_size), fPermissions(prot),
-							fNewAddress(0), fMappedAddress(NULL) {
+							fSize(vmsize), fFileOffset(offset), fFileSize(file_size), fAlignment(align),
+							fPermissions(prot), fSectionCount(sectionCount), fSectionsSize(sectionsSize),
+							fSectionsAlignment(sectionsAlignment), fNewAddress(0), fMappedAddress(NULL) {
 								strlcpy(fOrigName, segName, 16);
 							}
 							
@@ -84,8 +88,12 @@ public:
 		bool		readable() const	{ return fPermissions & VM_PROT_READ; }
 		bool		writable() const	{ return fPermissions & VM_PROT_WRITE; }
 		bool		executable() const	{ return fPermissions & VM_PROT_EXECUTE; }
+		uint64_t	alignment() const	{ return fAlignment; }
 		const char* name() const		{ return fOrigName; }
 		uint64_t	newAddress() const	{ return fNewAddress; }
+		uint32_t	sectionCount() const{ return fSectionCount; }
+		uint64_t	sectionsSize() const{ return fSectionsSize; }
+		uint64_t	sectionsAlignment() const { return fSectionsAlignment; }
 		void*		mappedAddress() const			{ return fMappedAddress; }
 		void		setNewAddress(uint64_t addr)	{ fNewAddress = addr; }
 		void		setMappedAddress(void* addr)	{ fMappedAddress = addr; }
@@ -93,6 +101,7 @@ public:
 		void		setFileOffset(uint64_t new_off)	{ fFileOffset = new_off; }
 		void		setFileSize(uint64_t new_size)	{ fFileSize = new_size; }
 		void		setWritable(bool w)				{ if (w) fPermissions |= VM_PROT_WRITE; else fPermissions &= ~VM_PROT_WRITE; }
+		void		setSectionsAlignment(uint64_t v){ fSectionsAlignment = v; }
 		void		reset()							{ fSize=fOrigSize; fFileOffset=fOrigFileOffset; fFileSize=fOrigFileSize; fPermissions=fOrigPermissions; }
 	private:
 		uint64_t		fOrigAddress;
@@ -104,7 +113,11 @@ public:
 		uint64_t		fSize;
 		uint64_t		fFileOffset;
 		uint64_t		fFileSize;
+		uint64_t		fAlignment;
 		uint32_t		fPermissions;
+		uint32_t		fSectionCount;
+		uint64_t		fSectionsSize;
+		uint64_t		fSectionsAlignment;
 		uint64_t		fNewAddress;
 		void*			fMappedAddress;
 	};
@@ -124,15 +137,23 @@ public:
 	virtual uint32_t							getFileType() const	= 0;
 	virtual uint32_t							getFlags() const = 0;
 	virtual	Library								getID() const = 0;
+	virtual bool								isDylib() const = 0;
 	virtual bool								isSplitSeg() const = 0;
 	virtual bool								hasSplitSegInfo() const = 0;
-	virtual bool								isRootOwned() const = 0;
+	virtual bool								hasSplitSegInfoV2() const = 0;
+	virtual int									notTrusted() const = 0;
 	virtual bool								inSharableLocation() const = 0;
+	virtual bool								hasDynamicLookupLinkage() const = 0;
+	virtual bool								hasMainExecutableLookupLinkage() const = 0;
+	virtual bool								isTwoLevelNamespace() const	= 0;
+	virtual bool								hasDyldInfo() const	= 0;
+	virtual bool								hasMultipleReadWriteSegments() const = 0;
 	virtual	uint32_t							getNameFileOffset() const = 0;
 	virtual time_t								getLastModTime() const = 0;
 	virtual ino_t								getInode() const = 0;
 	virtual std::vector<Segment>&				getSegments() = 0;
 	virtual const std::vector<Segment>&			getSegments() const = 0;
+	virtual const Segment*						getSegment(const char* name) const = 0;
 	virtual const std::vector<Library>&			getLibraries() const = 0;
 	virtual uint64_t							getBaseAddress() const = 0;
 	virtual uint64_t							getVMSize() const = 0;
@@ -142,6 +163,10 @@ public:
 	virtual uint64_t							getExecutableVMSize() const = 0;
 	virtual uint64_t							getWritableVMSize() const = 0;
 	virtual uint64_t							getReadOnlyVMSize() const = 0;
+	// need getDyldInfoExports because export info uses ULEB encoding and size could grow
+	virtual const uint8_t*						getDyldInfoExports() const = 0;
+	virtual void								setDyldInfoExports(const uint8_t* newExports) const = 0;
+	virtual void								uuid(uuid_t u) const = 0;
 };
 
 
@@ -161,15 +186,23 @@ public:
 	virtual uint32_t							getFileType() const		{ return fFileType; }
 	virtual uint32_t							getFlags() const		{ return fFlags; }
 	virtual	Library								getID() const			{ return fDylibID; }
+	virtual bool								isDylib() const			{ return fIsDylib; }
 	virtual bool								isSplitSeg() const;
-	virtual bool								hasSplitSegInfo() const	{ return fHasSplitSegInfo; }
-	virtual bool								isRootOwned() const		{ return fRootOwned; }
+	virtual bool								hasSplitSegInfo() const	{ return fSplitSegInfo != NULL; }
+	virtual bool								hasSplitSegInfoV2() const{ return fHasSplitSegInfoV2; }
+	virtual int									notTrusted() const		{ return fRootlessErrno; }
 	virtual bool								inSharableLocation() const { return fShareableLocation; }
+	virtual bool								hasDynamicLookupLinkage() const { return fDynamicLookupLinkage; }
+	virtual bool								hasMainExecutableLookupLinkage() const { return fMainExecutableLookupLinkage; }
+	virtual bool								isTwoLevelNamespace() const	{ return (fFlags & MH_TWOLEVEL); }
+	virtual bool								hasDyldInfo() const		{ return fHasDyldInfo; }
+	virtual bool								hasMultipleReadWriteSegments() const { return fHasTooManyWritableSegments; }
 	virtual	uint32_t							getNameFileOffset() const{ return fNameFileOffset; }
 	virtual time_t								getLastModTime() const	{ return fMTime; }
 	virtual ino_t								getInode() const		{ return fInode; }
 	virtual std::vector<Segment>&				getSegments()			{ return fSegments; }
 	virtual const std::vector<Segment>&			getSegments() const		{ return fSegments; }
+	virtual const Segment*						getSegment(const char* name) const;
 	virtual const std::vector<Library>&			getLibraries() const	{ return fLibraries; }
 	virtual uint64_t							getBaseAddress() const	{ return fLowSegment->address(); }
 	virtual uint64_t							getVMSize() const		{ return fVMSize; }
@@ -179,12 +212,23 @@ public:
 	virtual uint64_t							getExecutableVMSize() const		{ return fVMExecutableSize; }
 	virtual uint64_t							getWritableVMSize() const		{ return fVMWritablSize; }
 	virtual uint64_t							getReadOnlyVMSize() const		{ return fVMReadOnlySize; }
+	virtual const uint8_t*						getDyldInfoExports() const		{ return fDyldInfoExports; }
+	virtual void								setDyldInfoExports(const uint8_t* newExports) const { fDyldInfoExports = newExports; }
+	virtual void								uuid(uuid_t u) const { memcpy(u, fUUID, 16); }
 	
 private:
 	typedef typename A::P					P;
 	typedef typename A::P::E				E;
 	typedef typename A::P::uint_t			pint_t;
 	
+	uint64_t									segmentSize(const macho_segment_command<typename A::P>* segCmd) const;
+	uint64_t									segmentFileSize(const macho_segment_command<typename A::P>* segCmd) const;
+	uint64_t									segmentAlignment(const macho_segment_command<typename A::P>* segCmd) const;
+	uint64_t									sectionsSize(const macho_segment_command<typename A::P>* segCmd) const;
+	uint64_t									sectionsAlignment(const macho_segment_command<typename A::P>* segCmd) const;
+
+	bool										validReadWriteSeg(const Segment& seg) const;
+	
 	static cpu_type_t							arch();
 
 	const char*									fPath;
@@ -206,9 +250,17 @@ private:
 	uint64_t									fVMExecutableSize;
 	uint64_t									fVMWritablSize;
 	uint64_t									fVMReadOnlySize;
-	bool										fHasSplitSegInfo;
-	bool										fRootOwned;
+	const macho_linkedit_data_command<P>*		fSplitSegInfo;
+	bool										fHasSplitSegInfoV2;
+	int											fRootlessErrno;
 	bool										fShareableLocation;
+	bool										fDynamicLookupLinkage;
+	bool										fMainExecutableLookupLinkage;
+	bool										fIsDylib;
+	bool										fHasDyldInfo;
+	bool										fHasTooManyWritableSegments;
+	mutable const uint8_t*						fDyldInfoExports;
+	uuid_t										fUUID;
 };
 
 
@@ -224,14 +276,21 @@ public:
 	const std::vector<MachOLayoutAbstraction*>&	allLayouts() const { return fLayouts; }
 
 private:
+	class CStringHash {
+	public:
+		size_t operator()(const char* __s) const {
+			size_t __h = 0;
+			for ( ; *__s; ++__s)
+				__h = 5 * __h + *__s;
+			return __h;
+		};
+	};
 	struct CStringEquals {
 		bool operator()(const char* left, const char* right) const { return (strcmp(left, right) == 0); }
 	};
-	typedef __gnu_cxx::hash_map<const char*, const UniversalMachOLayout*, __gnu_cxx::hash<const char*>, CStringEquals> PathToNode;
+	typedef std::unordered_map<const char*, const UniversalMachOLayout*, CStringHash, CStringEquals> PathToNode;
 
-	static bool					compatibleSubtype(const std::set<ArchPair>* onlyArchs, cpu_type_t cpuType, cpu_subtype_t cpuSubType);
-	static bool					bestSliceForArch(uint32_t sliceCount, const struct fat_arch* slices, ArchPair ap, uint32_t& bestSliceIndex);
-	static const cpu_subtype_t* getArmSubtypeList(cpu_subtype_t s);
+	static bool					requestedSlice(const std::set<ArchPair>* onlyArchs, cpu_type_t cpuType, cpu_subtype_t cpuSubType);
 
 	static PathToNode							fgLayoutCache;
 	const char*									fPath;
@@ -243,80 +302,29 @@ UniversalMachOLayout::PathToNode UniversalMachOLayout::fgLayoutCache;
 
 
 
-// armv7 can run: v7, v6, v5, and v4
-static const cpu_subtype_t kARMV7compatibleSubTypes[] =
-	{ CPU_SUBTYPE_ARM_V7, CPU_SUBTYPE_ARM_V6, CPU_SUBTYPE_ARM_V5TEJ, CPU_SUBTYPE_ARM_V4T, CPU_SUBTYPE_ARM_ALL, 0 };
-
-// armv6 can run: v6, v5, and v4
-static const cpu_subtype_t kARMV6compatibleSubTypes[] =
-	{ CPU_SUBTYPE_ARM_V6, CPU_SUBTYPE_ARM_V5TEJ, CPU_SUBTYPE_ARM_V4T, CPU_SUBTYPE_ARM_ALL, 0};
-
-// xscale can run: xscale, v5, and v4
-static const cpu_subtype_t kARMXscaleCompatibleSubTypes[] =
-	{ CPU_SUBTYPE_ARM_XSCALE, CPU_SUBTYPE_ARM_V5TEJ, CPU_SUBTYPE_ARM_V4T, CPU_SUBTYPE_ARM_ALL, 0};
-
-// armv5 can run: v5 and v4
-static const cpu_subtype_t kARMV5compatibleSubTypes[] =
-	{ CPU_SUBTYPE_ARM_V5TEJ, CPU_SUBTYPE_ARM_V4T, CPU_SUBTYPE_ARM_ALL, 0};
-
-// armv4 can run: v4
-static const cpu_subtype_t kARMV4compatibleSubTypes[] =
-	{ CPU_SUBTYPE_ARM_V4T, CPU_SUBTYPE_ARM_ALL, 0 };
-
-const cpu_subtype_t* UniversalMachOLayout::getArmSubtypeList(cpu_subtype_t s)
-{
-	switch ( s ) {
-		case CPU_SUBTYPE_ARM_V7:
-			return kARMV7compatibleSubTypes;
-		case CPU_SUBTYPE_ARM_V6:
-			return kARMV6compatibleSubTypes;
-		case CPU_SUBTYPE_ARM_XSCALE:
-			return kARMXscaleCompatibleSubTypes;
-		case CPU_SUBTYPE_ARM_V5TEJ:
-			return kARMV5compatibleSubTypes;
-		case CPU_SUBTYPE_ARM_V4T:
-			return kARMV4compatibleSubTypes;
-	}
-	return NULL;
-}
-
-
-
 const MachOLayoutAbstraction* UniversalMachOLayout::getSlice(ArchPair ap) const
 {
-	switch ( ap.arch ) {
-		case CPU_TYPE_POWERPC:
-		case CPU_TYPE_I386:
-		case CPU_TYPE_X86_64:
-			// use first matching cputype
-			for(std::vector<MachOLayoutAbstraction*>::const_iterator it=fLayouts.begin(); it != fLayouts.end(); ++it) {
-				const MachOLayoutAbstraction* layout = *it;
-				if ( layout->getArchPair().arch == ap.arch ) 
-					return layout;
-			}
-			break;
-		case CPU_TYPE_ARM:
-			const cpu_subtype_t* list = getArmSubtypeList(ap.subtype);
-			if ( list != NULL ) {
-				// known subtype, find best match
-				for(const cpu_subtype_t* s=list; *s != 0; ++s) {
-					for(std::vector<MachOLayoutAbstraction*>::const_iterator it=fLayouts.begin(); it != fLayouts.end(); ++it) {
-						const MachOLayoutAbstraction* layout = *it;
-						if ( (layout->getArchPair().arch == ap.arch) && (layout->getArchPair().subtype == *s) ) 
-							return layout;
-					}
-				}
-			}
-			else {
-				// unknown arm sub-type, must have exact match
-				for(std::vector<MachOLayoutAbstraction*>::const_iterator it=fLayouts.begin(); it != fLayouts.end(); ++it) {
-					const MachOLayoutAbstraction* layout = *it;
-					if ( (layout->getArchPair().arch == ap.arch) && (layout->getArchPair().subtype == ap.subtype) ) 
-						return layout;
-				}
-			}
+	// use matching cputype and cpusubtype
+	for(std::vector<MachOLayoutAbstraction*>::const_iterator it=fLayouts.begin(); it != fLayouts.end(); ++it) {
+		const MachOLayoutAbstraction* layout = *it;
+		if ( layout->getArchPair().arch == ap.arch ) {
+            switch ( ap.arch ) {
+                case CPU_TYPE_ARM:
+				case CPU_TYPE_X86_64:
+                   if ( (layout->getArchPair().subtype & ~CPU_SUBTYPE_MASK) == (ap.subtype & ~CPU_SUBTYPE_MASK) )
+                        return layout;
+                    break;
+                 default:
+                    return layout;
+            }
+        }
 	}
-	throwf("no compatible slice found in %s", fPath);
+	// if requesting x86_64h and it did not exist, try x86_64 as a fallback
+	if ((ap.arch == CPU_TYPE_X86_64) && (ap.subtype == CPU_SUBTYPE_X86_64_H)) {
+		ap.subtype = CPU_SUBTYPE_X86_64_ALL;
+		return this->getSlice(ap);
+	}
+	return NULL;
 }
 
 
@@ -336,74 +344,24 @@ const UniversalMachOLayout& UniversalMachOLayout::find(const char* path, const s
 	return *result;
 }
 
-bool UniversalMachOLayout::bestSliceForArch(uint32_t sliceCount, const struct fat_arch* slices, ArchPair ap, uint32_t& bestSliceIndex)
-{
-	switch ( ap.arch ) {
-		case CPU_TYPE_POWERPC:
-		case CPU_TYPE_I386:
-		case CPU_TYPE_X86_64:
-			// use first matching cputype
-			for (uint32_t i=0; i < sliceCount; ++i) {
-				if ( OSSwapBigToHostInt32(slices[i].cputype) == ap.arch ) {
-					bestSliceIndex = i;
-					return true;
-				}
-			}
-			return false;
-		case CPU_TYPE_ARM:
-			// find best matching arch
-			const cpu_subtype_t* list = getArmSubtypeList(ap.subtype);
-			if ( list != NULL ) {
-				for(const cpu_subtype_t* s=list; *s != 0; ++s) {
-					for (uint32_t i=0; i < sliceCount; ++i) {
-						if ( (OSSwapBigToHostInt32(slices[i].cputype) == ap.arch) && (OSSwapBigToHostInt32(slices[i].cpusubtype) == *s) ) {
-							bestSliceIndex = i;
-							return true;
-						}
-					}
-				}
-				return false;
-			}
-			// unknown arm sub-type, must have exact match
-			for (uint32_t i=0; i < sliceCount; ++i) {
-				if ( (OSSwapBigToHostInt32(slices[i].cputype) == ap.arch) && (OSSwapBigToHostInt32(slices[i].cpusubtype) == ap.subtype) ) {
-					bestSliceIndex = i;
-					return true;
-				}
-			}
-			return false;
-	}
-	throw "unknown architecture";
-}
 
-bool UniversalMachOLayout::compatibleSubtype(const std::set<ArchPair>* onlyArchs, cpu_type_t cpuType, cpu_subtype_t cpuSubType)
+bool UniversalMachOLayout::requestedSlice(const std::set<ArchPair>* onlyArchs, cpu_type_t cpuType, cpu_subtype_t cpuSubType)
 {
+	if ( onlyArchs == NULL )
+		return true;
+	// must match cputype and cpusubtype
 	for (std::set<ArchPair>::const_iterator it = onlyArchs->begin(); it != onlyArchs->end(); ++it) {
-		if ( cpuType == it->arch ) {
-			switch ( it->arch ) {
-				case CPU_TYPE_POWERPC:
-				case CPU_TYPE_I386:
-				case CPU_TYPE_X86_64:
-					// just match cpu type
-					return true;
-				case CPU_TYPE_ARM:
-				{
-					const cpu_subtype_t* list = getArmSubtypeList(it->subtype);
-					if ( list != NULL ) {
-						// see if mach-o file is supported by this ArchPair
-						for(const cpu_subtype_t* s=list; *s != 0; ++s) {
-							if ( *s == cpuSubType )
-								return true;
-						}
-					}
-					else {
-						// unknown arm sub-type, must have exact match
-						if ( it->subtype == cpuSubType )
-							return true;
-					}
-				}
-			}
-		}
+		ArchPair anArch = *it;
+		if ( cpuType == anArch.arch ) {
+            switch ( cpuType ) {
+                case CPU_TYPE_ARM:
+                    if ( cpuSubType == anArch.subtype ) 
+                        return true;
+                    break;
+                default:
+                    return true;
+            }
+        }
 	}
 	return false;
 }
@@ -414,8 +372,13 @@ UniversalMachOLayout::UniversalMachOLayout(const char* path, const std::set<Arch
 {
 	// map in whole file
 	int fd = ::open(path, O_RDONLY, 0);
-	if ( fd == -1 )
-		throwf("can't open file, errno=%d", errno);
+	if ( fd == -1 ) {
+		int err = errno;
+		if  ( err == ENOENT )
+			throwf("file not found");
+		else
+			throwf("can't open file, errno=%d", err);
+	}
 	struct stat stat_buf;
 	if ( fstat(fd, &stat_buf) == -1)
 		throwf("can't stat open file %s, errno=%d", path, errno);
@@ -434,32 +397,19 @@ UniversalMachOLayout::UniversalMachOLayout(const char* path, const std::set<Arch
 			// Fat header is always big-endian
 			const struct fat_arch* slices = (struct fat_arch*)(p + sizeof(struct fat_header));
 			const uint32_t sliceCount = OSSwapBigToHostInt32(fh->nfat_arch);
-			std::set<uint32_t> slicesToUse;
-			if ( onlyArchs == NULL ) {
-				// no filter, so instantiate all slices
-				for (uint32_t i=0; i < sliceCount; ++i) 
-					slicesToUse.insert(i);
-			}
-			else {
-				// instantiate only slices that are best for each architecture
-				for (std::set<ArchPair>::const_iterator it = onlyArchs->begin(); it != onlyArchs->end(); ++it) {
-					uint32_t bestSliceIndex;
-					if ( bestSliceForArch(sliceCount, slices, *it, bestSliceIndex) ) 
-						slicesToUse.insert(bestSliceIndex);
-				}
-			}
 			for (uint32_t i=0; i < sliceCount; ++i) {
-				if ( slicesToUse.count(i) ) {
+				if ( requestedSlice(onlyArchs, OSSwapBigToHostInt32(slices[i].cputype), OSSwapBigToHostInt32(slices[i].cpusubtype)) ) {
 					uint32_t fileOffset = OSSwapBigToHostInt32(slices[i].offset);
 					if ( fileOffset > stat_buf.st_size ) {
 						throwf("malformed universal file, slice %u for architecture 0x%08X is beyond end of file: %s", 
 								i, OSSwapBigToHostInt32(slices[i].cputype), path);
 					}
+					if ( (fileOffset+OSSwapBigToHostInt32(slices[i].size)) > stat_buf.st_size ) {
+						throwf("malformed universal file, slice %u for architecture 0x%08X is beyond end of file: %s", 
+								i, OSSwapBigToHostInt32(slices[i].cputype), path);
+					}
 					try {
 						switch ( OSSwapBigToHostInt32(slices[i].cputype) ) {
-							case CPU_TYPE_POWERPC:
-								fLayouts.push_back(new MachOLayout<ppc>(&p[fileOffset], fileOffset, fPath, stat_buf.st_ino, stat_buf.st_mtime, stat_buf.st_uid));
-								break;
 							case CPU_TYPE_I386:
 								fLayouts.push_back(new MachOLayout<x86>(&p[fileOffset], fileOffset, fPath, stat_buf.st_ino, stat_buf.st_mtime, stat_buf.st_uid));
 								break;
@@ -469,8 +419,8 @@ UniversalMachOLayout::UniversalMachOLayout(const char* path, const std::set<Arch
 							case CPU_TYPE_ARM:
 								fLayouts.push_back(new MachOLayout<arm>(&p[fileOffset], fileOffset, fPath, stat_buf.st_ino, stat_buf.st_mtime, stat_buf.st_uid));
 								break;
-							case CPU_TYPE_POWERPC64:
-								// ignore ppc64 slices
+							case CPU_TYPE_ARM64:
+								fLayouts.push_back(new MachOLayout<arm64>(&p[fileOffset], fileOffset, fPath, stat_buf.st_ino, stat_buf.st_mtime, stat_buf.st_uid));
 								break;
 							default:
 								throw "unknown slice in fat file";
@@ -484,24 +434,21 @@ UniversalMachOLayout::UniversalMachOLayout(const char* path, const std::set<Arch
 		}
 		else {
 			try {
-				if ( (OSSwapBigToHostInt32(mh->magic) == MH_MAGIC) && (OSSwapBigToHostInt32(mh->cputype) == CPU_TYPE_POWERPC)) {
-					if ( (onlyArchs == NULL) || compatibleSubtype(onlyArchs, OSSwapLittleToHostInt32(mh->cputype), OSSwapLittleToHostInt32(mh->cpusubtype)) ) 
-						fLayouts.push_back(new MachOLayout<ppc>(mh, 0, fPath, stat_buf.st_ino, stat_buf.st_mtime, stat_buf.st_uid));
-				}
-				else if ( (OSSwapLittleToHostInt32(mh->magic) == MH_MAGIC) && (OSSwapLittleToHostInt32(mh->cputype) == CPU_TYPE_I386)) {
-					if ( (onlyArchs == NULL) || compatibleSubtype(onlyArchs, OSSwapLittleToHostInt32(mh->cputype), OSSwapLittleToHostInt32(mh->cpusubtype)) ) 
+				if ( (OSSwapLittleToHostInt32(mh->magic) == MH_MAGIC) && (OSSwapLittleToHostInt32(mh->cputype) == CPU_TYPE_I386)) {
+					if ( requestedSlice(onlyArchs, OSSwapLittleToHostInt32(mh->cputype), OSSwapLittleToHostInt32(mh->cpusubtype)) ) 
 						fLayouts.push_back(new MachOLayout<x86>(mh, 0, fPath, stat_buf.st_ino, stat_buf.st_mtime, stat_buf.st_uid));
 				}
 				else if ( (OSSwapLittleToHostInt32(mh->magic) == MH_MAGIC_64) && (OSSwapLittleToHostInt32(mh->cputype) == CPU_TYPE_X86_64)) {
-					if ( (onlyArchs == NULL) || compatibleSubtype(onlyArchs, OSSwapLittleToHostInt32(mh->cputype), OSSwapLittleToHostInt32(mh->cpusubtype)) ) 
+					if ( requestedSlice(onlyArchs, OSSwapLittleToHostInt32(mh->cputype), OSSwapLittleToHostInt32(mh->cpusubtype)) ) 
 						fLayouts.push_back(new MachOLayout<x86_64>(mh, 0, fPath, stat_buf.st_ino, stat_buf.st_mtime, stat_buf.st_uid));
 				}
 				else if ( (OSSwapLittleToHostInt32(mh->magic) == MH_MAGIC) && (OSSwapLittleToHostInt32(mh->cputype) == CPU_TYPE_ARM)) {
-					if ( (onlyArchs == NULL) || compatibleSubtype(onlyArchs, OSSwapLittleToHostInt32(mh->cputype), OSSwapLittleToHostInt32(mh->cpusubtype)) ) 
+					if ( requestedSlice(onlyArchs, OSSwapLittleToHostInt32(mh->cputype), OSSwapLittleToHostInt32(mh->cpusubtype)) ) 
 						fLayouts.push_back(new MachOLayout<arm>(mh, 0, fPath, stat_buf.st_ino, stat_buf.st_mtime, stat_buf.st_uid));
 				}
-				else if ( (OSSwapBigToHostInt32(mh->magic) == MH_MAGIC_64) && (OSSwapBigToHostInt32(mh->cputype) == CPU_TYPE_POWERPC64)) {
-					// ignore ppc64 slices
+				else if ( (OSSwapLittleToHostInt32(mh->magic) == MH_MAGIC_64) && (OSSwapLittleToHostInt32(mh->cputype) == CPU_TYPE_ARM64)) {
+					if ( requestedSlice(onlyArchs, OSSwapLittleToHostInt32(mh->cputype), OSSwapLittleToHostInt32(mh->cpusubtype)) ) 
+						fLayouts.push_back(new MachOLayout<arm64>(mh, 0, fPath, stat_buf.st_ino, stat_buf.st_mtime, stat_buf.st_uid));
 				}
 				else {
 					throw "unknown file format";
@@ -519,20 +466,112 @@ UniversalMachOLayout::UniversalMachOLayout(const char* path, const std::set<Arch
 }
 
 
+template <typename A>
+uint64_t MachOLayout<A>::segmentSize(const macho_segment_command<typename A::P>* segCmd) const
+{
+	// <rdar://problem/13089366> segments may have 16KB alignment padding at end, if so we can remove that in cache
+	if ( segCmd->nsects() > 0 ) {
+		const macho_section<P>* const sectionsStart = (macho_section<P>*)((uint8_t*)segCmd + sizeof(macho_segment_command<P>));
+		const macho_section<P>* const lastSection = &sectionsStart[segCmd->nsects()-1];
+		uint64_t endSectAddr = lastSection->addr() + lastSection->size();
+		uint64_t endSectAddrPage = (endSectAddr + 4095) & (-4096);
+		if ( endSectAddrPage < (segCmd->vmaddr() + segCmd->vmsize()) ) {
+			uint64_t size =  endSectAddrPage - segCmd->vmaddr();
+			//if ( size != segCmd->vmsize() )
+			//	fprintf(stderr, "trim %s size=0x%08llX instead of 0x%08llX for %s\n", 
+			//		segCmd->segname(), size, segCmd->vmsize(), getFilePath());
+			return size;
+		}
+	}
+	return segCmd->vmsize();
+}
+
+template <typename A>
+uint64_t MachOLayout<A>::segmentAlignment(const macho_segment_command<typename A::P>* segCmd) const
+{
+	int p2align = 12;
+	if ( segCmd->nsects() > 0 ) {
+		const macho_section<P>* const sectionsStart = (macho_section<P>*)((uint8_t*)segCmd + sizeof(macho_segment_command<P>));
+		const macho_section<P>* const sectionsEnd = &sectionsStart[segCmd->nsects()-1];
+		for (const macho_section<P>*  sect=sectionsStart; sect < sectionsEnd; ++sect) {
+			if ( sect->align() > p2align )
+				p2align = sect->align();
+		}
+	}
+	return (1 << p2align);
+}
+
+template <typename A>
+uint64_t MachOLayout<A>::segmentFileSize(const macho_segment_command<typename A::P>* segCmd) const
+{
+	// <rdar://problem/13089366> segments may have 16KB alignment padding at end, if so we can remove that in cache
+	if ( segCmd->nsects() > 0 ) {
+		uint64_t endOffset = segCmd->fileoff();
+		const macho_section<P>* const sectionsStart = (macho_section<P>*)((uint8_t*)segCmd + sizeof(macho_segment_command<P>));
+		const macho_section<P>* const sectionsEnd = &sectionsStart[segCmd->nsects()];
+		for (const macho_section<P>* sect=sectionsStart; sect < sectionsEnd; ++sect) {
+			if ( sect->offset() != 0 )
+				endOffset = sect->offset() + sect->size();
+		}
+		uint64_t size = (endOffset - segCmd->fileoff() + 4095) & (-4096);
+		//if ( size != segCmd->filesize() )
+		//	fprintf(stderr, "trim %s filesize=0x%08llX instead of 0x%08llX for %s\n", 
+		//		segCmd->segname(), size, segCmd->filesize(), getFilePath());
+		return size;
+	}
+	return segCmd->filesize();
+}
+
+template <typename A>
+uint64_t MachOLayout<A>::sectionsSize(const macho_segment_command<typename A::P>* segCmd) const
+{
+	if ( segCmd->nsects() > 0 ) {
+		const macho_section<P>* const sectionsStart = (macho_section<P>*)((uint8_t*)segCmd + sizeof(macho_segment_command<P>));
+		const macho_section<P>* const lastSection = &sectionsStart[segCmd->nsects()-1];
+		uint64_t endSectAddr = lastSection->addr() + lastSection->size();
+		if ( endSectAddr < (segCmd->vmaddr() + segCmd->vmsize()) ) {
+			uint64_t size =  endSectAddr - segCmd->vmaddr();
+			return size;
+		}
+	}
+	return segCmd->vmsize();
+}
+
+template <typename A>
+uint64_t MachOLayout<A>::sectionsAlignment(const macho_segment_command<typename A::P>* segCmd) const
+{
+	int p2align = 4;
+	if ( hasSplitSegInfoV2() && (segCmd->nsects() > 0) ) {
+		const macho_section<P>* const sectionsStart = (macho_section<P>*)((uint8_t*)segCmd + sizeof(macho_segment_command<P>));
+		const macho_section<P>* const sectionsEnd = &sectionsStart[segCmd->nsects()-1];
+		for (const macho_section<P>*  sect=sectionsStart; sect < sectionsEnd; ++sect) {
+			if ( sect->align() > p2align )
+				p2align = sect->align();
+		}
+	}
+	return (1 << p2align);
+}
+
+
+
 template <typename A>
 MachOLayout<A>::MachOLayout(const void* machHeader, uint64_t offset, const char* path, ino_t inode, time_t modTime, uid_t uid)
- : fPath(path), fOffset(offset), fArchPair(0,0), fMTime(modTime), fInode(inode), fHasSplitSegInfo(false), fRootOwned(uid==0),
-   fShareableLocation(false)
+ : fPath(path), fOffset(offset), fArchPair(0,0), fMTime(modTime), fInode(inode), fSplitSegInfo(NULL), fHasSplitSegInfoV2(false), fRootlessErrno(0),
+   fShareableLocation(false), fDynamicLookupLinkage(false), fMainExecutableLookupLinkage(false), fIsDylib(false), 
+	fHasDyldInfo(false), fHasTooManyWritableSegments(false), fDyldInfoExports(NULL)
 {
 	fDylibID.name = NULL;
 	fDylibID.currentVersion = 0;
 	fDylibID.compatibilityVersion = 0;
-
+	bzero(fUUID, sizeof(fUUID));
+	
 	const macho_header<P>* mh = (const macho_header<P>*)machHeader;
 	if ( mh->cputype() != arch() )
 		throw "Layout object is wrong architecture";
 	switch ( mh->filetype() ) {
 		case MH_DYLIB:
+			fIsDylib = true;
+			break;
 		case MH_BUNDLE:
 		case MH_EXECUTE:
 		case MH_DYLIB_STUB:
@@ -545,7 +584,12 @@ MachOLayout<A>::MachOLayout(const void* machHeader, uint64_t offset, const char*
 	fFileType = mh->filetype();
 	fArchPair.arch = mh->cputype();
 	fArchPair.subtype = mh->cpusubtype();
-	
+	if ( rootless_check_trusted(path) != 0 && rootless_protected_volume(path) == 1)
+		fRootlessErrno = errno;
+
+	const macho_dyld_info_command<P>* dyldInfo = NULL;
+	const macho_symtab_command<P>* symbolTableCmd = NULL;
+	const macho_dysymtab_command<P>* dynamicSymbolTableCmd = NULL;
 	const macho_load_command<P>* const cmds = (macho_load_command<P>*)((uint8_t*)mh + sizeof(macho_header<P>));
 	const uint32_t cmd_count = mh->ncmds();
 	const macho_load_command<P>* cmd = cmds;
@@ -564,6 +608,7 @@ MachOLayout<A>::MachOLayout(const void* machHeader, uint64_t offset, const char*
 			case LC_LOAD_DYLIB:
 			case LC_LOAD_WEAK_DYLIB:
 			case LC_REEXPORT_DYLIB:
+			case LC_LOAD_UPWARD_DYLIB:
 				{
 					macho_dylib_command<P>* dylib = (macho_dylib_command<P>*)cmd;
 					Library lib;
@@ -575,13 +620,32 @@ MachOLayout<A>::MachOLayout(const void* machHeader, uint64_t offset, const char*
 				}
 				break;
 			case LC_SEGMENT_SPLIT_INFO:
-				fHasSplitSegInfo = true;
+				fSplitSegInfo = (macho_linkedit_data_command<P>*)cmd;
 				break;
 			case macho_segment_command<P>::CMD:
 				{
-					macho_segment_command<P>* segCmd = (macho_segment_command<P>*)cmd;
-					fSegments.push_back(Segment(segCmd->vmaddr(), segCmd->vmsize(), segCmd->fileoff(), 
-								segCmd->filesize(), segCmd->initprot(), segCmd->segname()));
+					const macho_segment_command<P>* segCmd = (macho_segment_command<P>*)cmd;
+					fSegments.push_back(Segment(segCmd->vmaddr(), segmentSize(segCmd), segCmd->fileoff(), 
+								segmentFileSize(segCmd), sectionsSize(segCmd), sectionsAlignment(segCmd),
+								segmentAlignment(segCmd), segCmd->initprot(),
+								segCmd->nsects(), segCmd->segname()));
+				}
+				break;
+			case LC_SYMTAB:
+				symbolTableCmd = (macho_symtab_command<P>*)cmd;
+				break;
+			case LC_DYSYMTAB:
+				dynamicSymbolTableCmd = (macho_dysymtab_command<P>*)cmd;
+				break;
+			case LC_DYLD_INFO:
+			case LC_DYLD_INFO_ONLY:
+				fHasDyldInfo = true;
+				dyldInfo = (struct macho_dyld_info_command<P>*)cmd;
+				break;
+			case LC_UUID:
+				{
+					const macho_uuid_command<P>* uc = (macho_uuid_command<P>*)cmd;
+					memcpy(&fUUID, uc->uuid(), 16);
 				}
 				break;
 		}
@@ -612,29 +676,71 @@ MachOLayout<A>::MachOLayout(const void* machHeader, uint64_t offset, const char*
 			if ( (fLowWritableSegment == NULL) || (seg.address() < fLowWritableSegment->address()) )
 				fLowWritableSegment = &seg;
 			fVMWritablSize += seg.size();
+			if ( !validReadWriteSeg(seg) ) {
+				fHasTooManyWritableSegments = true;
+			}
 		}
 		else {
 			if ( (fLowReadOnlySegment == NULL) || (seg.address() < fLowReadOnlySegment->address()) )
 				fLowReadOnlySegment = &seg;
 			fVMReadOnlySize += seg.size();
-		}
+		}		
 	}
 	if ( (highSegment != NULL) && (fLowSegment != NULL) )
 		fVMSize = (highSegment->address() + highSegment->size() - fLowSegment->address() + 4095) & (-4096);			
+
+	// scan undefines looking, for magic ordinals
+	if ( (symbolTableCmd != NULL) && (dynamicSymbolTableCmd != NULL) ) {
+		const macho_nlist<P>* symbolTable = (macho_nlist<P>*)((uint8_t*)machHeader + symbolTableCmd->symoff());
+		const uint32_t startUndefs = dynamicSymbolTableCmd->iundefsym();
+		const uint32_t endUndefs = startUndefs + dynamicSymbolTableCmd->nundefsym();
+		for (uint32_t i=startUndefs; i < endUndefs; ++i) {
+			uint8_t ordinal = GET_LIBRARY_ORDINAL(symbolTable[i].n_desc());
+			if ( ordinal == DYNAMIC_LOOKUP_ORDINAL )
+				fDynamicLookupLinkage = true;
+			else if ( ordinal == EXECUTABLE_ORDINAL )
+				fMainExecutableLookupLinkage = true;
+		}
+	}
+	
+	if ( dyldInfo != NULL ) {
+		if ( dyldInfo->export_off() != 0 ) {
+			fDyldInfoExports = (uint8_t*)machHeader + dyldInfo->export_off();
+		}
+	}
+
+	if ( fSplitSegInfo != NULL ) {
+		const uint8_t* infoStart = (uint8_t*)machHeader + fSplitSegInfo->dataoff();
+		fHasSplitSegInfoV2 = ( *infoStart == DYLD_CACHE_ADJ_V2_FORMAT );
+		if ( !fHasSplitSegInfoV2 ) {
+			// split seg version not known when segments created
+			// v1 does not support packing, so simulate that by forcing alignment
+			for (Segment& seg : fSegments) {
+				seg.setSectionsAlignment(4096);
+			}
+		}
+	}
+
 }
 
-template <> cpu_type_t MachOLayout<ppc>::arch()     { return CPU_TYPE_POWERPC; }
 template <> cpu_type_t MachOLayout<x86>::arch()     { return CPU_TYPE_I386; }
 template <> cpu_type_t MachOLayout<x86_64>::arch()  { return CPU_TYPE_X86_64; }
 template <> cpu_type_t MachOLayout<arm>::arch()		{ return CPU_TYPE_ARM; }
-
+template <> cpu_type_t MachOLayout<arm64>::arch()	{ return CPU_TYPE_ARM64; }
 
 template <>
-bool MachOLayout<ppc>::isSplitSeg() const
+bool MachOLayout<x86>::validReadWriteSeg(const Segment& seg) const
 {
-	return ( (this->getFlags() & MH_SPLIT_SEGS) != 0 );
+	return (strcmp(seg.name(), "__DATA") == 0) || (strcmp(seg.name(), "__OBJC") == 0);
 }
 
+template <typename A>
+bool MachOLayout<A>::validReadWriteSeg(const Segment& seg) const
+{
+	return (strcmp(seg.name(), "__DATA") == 0);
+}
+
+
 template <>
 bool MachOLayout<x86>::isSplitSeg() const
 {
@@ -653,6 +759,18 @@ bool MachOLayout<A>::isSplitSeg() const
 	return false;
 }
 
+template <typename A>
+const MachOLayoutAbstraction::Segment* MachOLayout<A>::getSegment(const char* name) const
+{
+	for(std::vector<Segment>::const_iterator it = fSegments.begin(); it != fSegments.end(); ++it) {
+		const Segment& seg = *it;
+		if ( strcmp(seg.name(), name) == 0 )
+			return &seg;
+	}
+	return NULL;
+}
+
+
 
 #endif // __MACHO_LAYOUT__
 
