@@ -40,8 +40,11 @@ int main()
 	
 	// execute leaks command on myself
 	char cmd[512];
-	sprintf(cmd, "leaks %u\n", getpid());
-	system(cmd);
+	int result = system(cmd);
+	if ( result == EXIT_SUCCESS )
+		PASS("rpath-dlopen-leak");
+	else
+		FAIL("rpath-dlopen-leak");
 	 	
 	return EXIT_SUCCESS;
 }
